uniform sampler2D diffuseTexture;

uniform vec2 texelSize;

uniform float R_fxaaSpanMax;
uniform float R_reduceMin;
uniform float R_reduceMul;

varying vec2 outUV;

void main()
{
	vec3 luma = vec3(0.29, 0.59, 0.12);
	float lumaTL = dot(texture2D(diffuseTexture, outUV + (vec2(-1.0, -1.0) * texelSize)).xyz, luma);
	float lumaTR = dot(texture2D(diffuseTexture, outUV + (vec2(1.0, -1.0) * texelSize)).xyz, luma);
	float lumaBL = dot(texture2D(diffuseTexture, outUV + (vec2(-1.0, 1.0) * texelSize)).xyz, luma);
	float lumaBR = dot(texture2D(diffuseTexture, outUV + (vec2(1.0, 1.0) * texelSize)).xyz, luma);
	float lumaM = dot(texture2D(diffuseTexture, outUV).xyz, luma);
	
	vec2 dir;
	dir.x = ((lumaTL + lumaTR) - (lumaBL + lumaBR));
	dir.y = ((lumaTL + lumaBL) - (lumaTR + lumaBR));
	
	// vec3 edge = vec3(0.5*(dir.x+dir.y));
	// gl_FragColor = vec4(edge,1.0);
	
	float dirReduce = max((lumaTL + lumaTR + lumaBL + lumaBR) * (R_reduceMul * 0.25), R_reduceMin);
	float temp = 1.0/(min(abs(dir.x), abs(dir.y)) + dirReduce);
	dir = clamp(dir * temp, vec2(-R_fxaaSpanMax, -R_fxaaSpanMax), vec2(R_fxaaSpanMax, R_fxaaSpanMax)) * texelSize;
	
	vec3 result1 = 0.5 * (
		texture2D(diffuseTexture, outUV + (dir * vec2(1.0/3.0 - 0.5))).xyz +
		texture2D(diffuseTexture, outUV + (dir * vec2(2.0/3.0 - 0.5))).xyz); 
	
	vec3 result2 = result1 * 0.5 + 0.25 * (
		texture2D(diffuseTexture, outUV + (dir * vec2(0.0/3.0 - 0.5))).xyz +
		texture2D(diffuseTexture, outUV + (dir * vec2(3.0/3.0 - 0.5))).xyz);
	
	float lumaMin = min(lumaM, min(min(lumaTL, lumaTR), min(lumaBL, lumaBR)));
	float lumaMax = max(lumaM, max(max(lumaTL, lumaTR), max(lumaBL, lumaBR)));
	float lumaResult2 = dot(luma, result2);
	
	if(lumaResult2 < lumaMin || lumaResult2 > lumaMax)
		gl_FragColor = vec4(result1, 1.0);
	else
		gl_FragColor = vec4(result2, 1.0);
	
}