#code "base bones multiTexturing noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec4 outColor;
attribute vec4 color;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outColor = color;
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec4 outColor;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	gl_FragColor = textureColor * diffuse * outColor;
}

#endcode
#code "base bones noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec4 outColor;
attribute vec4 color;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	gl_FragColor = textureColor * diffuse * outColor;
}

#endcode
#code "base multiTexturing noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec4 outColor;
attribute vec4 color;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outColor = color;
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec4 outColor;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	gl_FragColor = textureColor * diffuse * outColor;
}

#endcode
#code "base noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec4 outColor;
attribute vec4 color;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	gl_FragColor = textureColor * diffuse * outColor;
}

#endcode
#code "ambientLighting base bones multiTexturing parallaxMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
uniform mat3 T_normalMatrix;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outPosition = vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 n = normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 b = tangent.w * cross(n, t);
	// t = normalize(t - (dot(t, n) * n)); //Ortho dinges
	tbnMatrix = mat3(t, b, n);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying mat3 tbnMatrix;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base bones parallaxMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
uniform mat3 T_normalMatrix;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outPosition = vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 n = normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 b = tangent.w * cross(n, t);
	// t = normalize(t - (dot(t, n) * n)); //Ortho dinges
	tbnMatrix = mat3(t, b, n);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying mat3 tbnMatrix;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base bones multiTexturing"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base bones"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
uniform sampler2D diffuseTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base multiTexturing parallaxMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
uniform mat3 T_normalMatrix;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outPosition = vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 n = normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 b = tangent.w * cross(n, t);
	// t = normalize(t - (dot(t, n) * n)); //Ortho dinges
	tbnMatrix = mat3(t, b, n);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying mat3 tbnMatrix;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base parallaxMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
uniform mat3 T_normalMatrix;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outPosition = vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 n = normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 b = tangent.w * cross(n, t);
	// t = normalize(t - (dot(t, n) * n)); //Ortho dinges
	tbnMatrix = mat3(t, b, n);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying mat3 tbnMatrix;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base multiTexturing"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "ambientLighting base"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_ambientLight;
uniform vec4 ambient;
uniform sampler2D diffuseTexture;
uniform sampler2D emissionTexture;
uniform vec4 emission;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 emissionColor = vec4(1.0); //output
	emissionColor = emission * texture2D(emissionTexture, uvCoords);
	gl_FragColor = textureColor * vec4(R_ambientLight,1.0) * ambient + vec4(emissionColor.rgb, 0.0);
}

#endcode
#code "bones fog instancedRendering"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
attribute mat4 instanceMatrix;
attribute vec4 instanceUvAnimation;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * instanceMatrix * vertexPosition;
	outUV = UV * instanceUvAnimation.zw + instanceUvAnimation.xy;
	outPosition=vec3(T_modelMatrix * instanceMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_fogColor;
uniform float R_fogBegin;
uniform float R_fogEnd;
uniform float R_fogDensity;
uniform int R_fogType;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	float dist = length(T_eyePos - outPosition);
	float fog = 0.0;
	if(R_fogType==0)
	{
		float fogWidth = R_fogEnd-R_fogBegin;
		fog = clamp((dist-R_fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(R_fogType==1)
	{
		fog = 1.0-clamp(exp(-R_fogDensity * (dist-R_fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((R_fogDensity * (dist-R_fogBegin)), 2.0)), 0.0, 0.9999);
	}
	float textureAlpha = texture2D(diffuseTexture, outUV).a;
	gl_FragColor = vec4(R_fogColor, 1.0) * fog * textureAlpha * diffuse.a;
}

#endcode
#code "bones fog"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_fogColor;
uniform float R_fogBegin;
uniform float R_fogEnd;
uniform float R_fogDensity;
uniform int R_fogType;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	float dist = length(T_eyePos - outPosition);
	float fog = 0.0;
	if(R_fogType==0)
	{
		float fogWidth = R_fogEnd-R_fogBegin;
		fog = clamp((dist-R_fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(R_fogType==1)
	{
		fog = 1.0-clamp(exp(-R_fogDensity * (dist-R_fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((R_fogDensity * (dist-R_fogBegin)), 2.0)), 0.0, 0.9999);
	}
	float textureAlpha = texture2D(diffuseTexture, outUV).a;
	gl_FragColor = vec4(R_fogColor, 1.0) * fog * textureAlpha * diffuse.a;
}

#endcode
#code "fog instancedRendering"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
attribute mat4 instanceMatrix;
attribute vec4 instanceUvAnimation;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * instanceMatrix * vertexPosition;
	outUV = UV * instanceUvAnimation.zw + instanceUvAnimation.xy;
	outPosition=vec3(T_modelMatrix * instanceMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_fogColor;
uniform float R_fogBegin;
uniform float R_fogEnd;
uniform float R_fogDensity;
uniform int R_fogType;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	float dist = length(T_eyePos - outPosition);
	float fog = 0.0;
	if(R_fogType==0)
	{
		float fogWidth = R_fogEnd-R_fogBegin;
		fog = clamp((dist-R_fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(R_fogType==1)
	{
		fog = 1.0-clamp(exp(-R_fogDensity * (dist-R_fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((R_fogDensity * (dist-R_fogBegin)), 2.0)), 0.0, 0.9999);
	}
	float textureAlpha = texture2D(diffuseTexture, outUV).a;
	gl_FragColor = vec4(R_fogColor, 1.0) * fog * textureAlpha * diffuse.a;
}

#endcode
#code "fog"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform vec3 R_fogColor;
uniform float R_fogBegin;
uniform float R_fogEnd;
uniform float R_fogDensity;
uniform int R_fogType;
uniform vec3 T_eyePos;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	float dist = length(T_eyePos - outPosition);
	float fog = 0.0;
	if(R_fogType==0)
	{
		float fogWidth = R_fogEnd-R_fogBegin;
		fog = clamp((dist-R_fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(R_fogType==1)
	{
		fog = 1.0-clamp(exp(-R_fogDensity * (dist-R_fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((R_fogDensity * (dist-R_fogBegin)), 2.0)), 0.0, 0.9999);
	}
	float textureAlpha = texture2D(diffuseTexture, outUV).a;
	gl_FragColor = vec4(R_fogColor, 1.0) * fog * textureAlpha * diffuse.a;
}

#endcode
#code "bones genShadows instancedRendering"
#shader vertex
uniform mat4 T_modelViewProjectionMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
attribute mat4 instanceMatrix;
attribute vec4 instanceUvAnimation;
varying vec2 outUV;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * instanceMatrix * vertexPosition;
	outUV = UV * instanceUvAnimation.zw + instanceUvAnimation.xy;
}

#shader fragment
varying vec2 outUV;
uniform float shadowAlphaThreshold;
uniform sampler2D diffuseTexture;

void main()
{
	float alpha = texture2D(diffuseTexture, outUV).a;
	if(alpha <= shadowAlphaThreshold) discard;
	float depth = gl_FragCoord.z;
	float dx = dFdx(depth);
	float dy = dFdy(depth);
	float moment2 = depth * depth + 0.25 * (dx * dx + dy * dy);
	gl_FragColor = vec4(depth, moment2, 0.0, 1.0);
}

#endcode
#code "bones genShadows"
#shader vertex
uniform mat4 T_modelViewProjectionMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec2 outUV;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
}

#shader fragment
varying vec2 outUV;
uniform float shadowAlphaThreshold;
uniform sampler2D diffuseTexture;

void main()
{
	float alpha = texture2D(diffuseTexture, outUV).a;
	if(alpha <= shadowAlphaThreshold) discard;
	float depth = gl_FragCoord.z;
	float dx = dFdx(depth);
	float dy = dFdy(depth);
	float moment2 = depth * depth + 0.25 * (dx * dx + dy * dy);
	gl_FragColor = vec4(depth, moment2, 0.0, 1.0);
}

#endcode
#code "genShadows instancedRendering"
#shader vertex
uniform mat4 T_modelViewProjectionMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
attribute mat4 instanceMatrix;
attribute vec4 instanceUvAnimation;
varying vec2 outUV;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * instanceMatrix * vertexPosition;
	outUV = UV * instanceUvAnimation.zw + instanceUvAnimation.xy;
}

#shader fragment
varying vec2 outUV;
uniform float shadowAlphaThreshold;
uniform sampler2D diffuseTexture;

void main()
{
	float alpha = texture2D(diffuseTexture, outUV).a;
	if(alpha <= shadowAlphaThreshold) discard;
	float depth = gl_FragCoord.z;
	float dx = dFdx(depth);
	float dy = dFdy(depth);
	float moment2 = depth * depth + 0.25 * (dx * dx + dy * dy);
	gl_FragColor = vec4(depth, moment2, 0.0, 1.0);
}

#endcode
#code "genShadows"
#shader vertex
uniform mat4 T_modelViewProjectionMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec2 outUV;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
}

#shader fragment
varying vec2 outUV;
uniform float shadowAlphaThreshold;
uniform sampler2D diffuseTexture;

void main()
{
	float alpha = texture2D(diffuseTexture, outUV).a;
	if(alpha <= shadowAlphaThreshold) discard;
	float depth = gl_FragCoord.z;
	float dx = dFdx(depth);
	float dy = dFdy(depth);
	float moment2 = depth * depth + 0.25 * (dx * dx + dy * dy);
	gl_FragColor = vec4(depth, moment2, 0.0, 1.0);
}

#endcode
