#code "environmentMapping fogDepthEffect noLighting reflection refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;
varying vec4 outRefractionPosition;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform vec3 fogColor;
uniform float fogBegin;
uniform float fogEnd;
uniform float fogDensity;
uniform int fogType;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float fog = 0.0;
	if(fogType==0)
	{
		float fogWidth = fogEnd-fogBegin;
		fog = clamp((waterDepth-fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(fogType==1)
	{
		fog = 1.0-clamp(exp(-fogDensity * (waterDepth-fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((fogDensity * (waterDepth-fogBegin)), 2.0)), 0.0, 0.9999);
	}
	refractionColor = mix(refractionColor, fogColor, fog);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "environmentMapping noLighting reflection refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;
varying vec4 outRefractionPosition;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "environmentMapping fogDepthEffect reflection refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;
varying vec4 outRefractionPosition;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform vec3 fogColor;
uniform float fogBegin;
uniform float fogEnd;
uniform float fogDensity;
uniform int fogType;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float fog = 0.0;
	if(fogType==0)
	{
		float fogWidth = fogEnd-fogBegin;
		fog = clamp((waterDepth-fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(fogType==1)
	{
		fog = 1.0-clamp(exp(-fogDensity * (waterDepth-fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((fogDensity * (waterDepth-fogBegin)), 2.0)), 0.0, 0.9999);
	}
	refractionColor = mix(refractionColor, fogColor, fog);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "environmentMapping reflection refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;
varying vec4 outRefractionPosition;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "environmentMapping fogDepthEffect noLighting refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outRefractionPosition;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform vec3 fogColor;
uniform float fogBegin;
uniform float fogEnd;
uniform float fogDensity;
uniform int fogType;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float fog = 0.0;
	if(fogType==0)
	{
		float fogWidth = fogEnd-fogBegin;
		fog = clamp((waterDepth-fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(fogType==1)
	{
		fog = 1.0-clamp(exp(-fogDensity * (waterDepth-fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((fogDensity * (waterDepth-fogBegin)), 2.0)), 0.0, 0.9999);
	}
	refractionColor = mix(refractionColor, fogColor, fog);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "environmentMapping noLighting refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outRefractionPosition;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "environmentMapping fogDepthEffect refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outRefractionPosition;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform vec3 fogColor;
uniform float fogBegin;
uniform float fogEnd;
uniform float fogDensity;
uniform int fogType;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float fog = 0.0;
	if(fogType==0)
	{
		float fogWidth = fogEnd-fogBegin;
		fog = clamp((waterDepth-fogBegin)/fogWidth, 0.0, 0.9999);
	} 
	else if(fogType==1)
	{
		fog = 1.0-clamp(exp(-fogDensity * (waterDepth-fogBegin)), 0.0, 0.9999);
	} 
	else
	{
		fog = 1.0-clamp(exp(-pow((fogDensity * (waterDepth-fogBegin)), 2.0)), 0.0, 0.9999);
	}
	refractionColor = mix(refractionColor, fogColor, fog);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "environmentMapping refraction"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outRefractionPosition;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outRefractionPosition = T_modelViewProjectionMatrix * vertexPosition;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform float alphaDepth;
uniform float distortionDepth;
uniform float R_near;
uniform float R_far;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outRefractionPosition;
uniform sampler2D R_refractionTexture;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//smooth edges
	float near = R_near;
	float far = R_far;
	float depth = texture2D(R_depthTexture, vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w)).r;
	float floorDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	depth = gl_FragCoord.z;
	float waterDistance = 2.0 * near * far / (far + near - (2.0 * depth - 1.0) * (far - near));
	float waterDepth = floorDistance - waterDistance;
	alpha = clamp(waterDepth / alphaDepth, 0.0, 1.0);
	distortionAlpha = clamp(waterDepth / distortionDepth, 0.0, 1.0);
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec2 refractionCoords = vec2(0.5) + 0.5 * (outRefractionPosition.xy / outRefractionPosition.w + perturbation);
	refractionColor = texture2D(R_refractionTexture, refractionCoords).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "environmentMapping noLighting reflection"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "environmentMapping reflection"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying vec4 outReflectedPosition;
uniform mat4 R_reflectionMatrix;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outReflectedPosition = R_reflectionMatrix * vertexPosition;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
uniform sampler2D dudvTexture;
uniform sampler2D R_depthTexture;
uniform float maxDistortionOffset;
uniform float distortionStrength;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
varying vec3 outNormal;
uniform vec3 T_eyePos;
varying vec4 outReflectedPosition;
uniform sampler2D R_reflectionTexture;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	float distortionAlpha = 1.0;
	//distortions
	distortedTexCoords = outUV + texture2D(dudvTexture, outUV * uvAnimation1.zw + uvAnimation1.xy).rg * maxDistortionOffset;
	vec2 perturbation = (texture2D(dudvTexture, distortedTexCoords * uvAnimation2.zw + uvAnimation2.xy).rg * 2.0 - 1.0) * distortionStrength * distortionAlpha;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec2 reflectionCoords = outReflectedPosition.xy / outReflectedPosition.w + perturbation;
	reflectionColor = texture2D(R_reflectionTexture, reflectionCoords).rgb;
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec2 uvCoords = outUV; //input
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeReflection cubeRefraction environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeReflection cubeRefraction environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeReflection cubeRefraction environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeReflection cubeRefraction environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeReflection cubeRefraction environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeReflection cubeRefraction environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeReflection cubeRefraction environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeReflection cubeRefraction environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float reflectivity;
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	float fresnelTerm = abs(dot(eyeVector, fresnelNormalVector));
	fresnelTerm = pow(fresnelTerm, reflectivity);
	combinedColor = mix(reflectionColor, refractionColor, fresnelTerm * refractionFactor);
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeReflection environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeReflection environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeReflection environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeReflection environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeReflection environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeReflection environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeReflection environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeReflection environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_reflectionTexture;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	vec3 refl = reflect(-eyeVector, normalVector);
	reflectionColor = texture2D(R_reflectionTexture, GetCubeTC(refl,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	float refractionFactor = 1.0;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeRefraction environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeRefraction environmentMapping noLighting normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeRefraction environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "cubeRefraction environmentMapping noLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
attribute vec4 color;
varying vec4 outColor;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	outColor = color;
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;
varying vec4 outColor;
uniform sampler2D diffuseTexture;
uniform vec4 diffuse;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
	vec3 diffuseLighting = (outColor * diffuse * texture2D(diffuseTexture, distortedTexCoords)).rgb;
	gl_FragColor += vec4(diffuseLighting, 0.0) * alpha;
}

#endcode
#code "bones cubeRefraction environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeRefraction environmentMapping normalMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
varying mat3 tbnMatrix;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	vec3 t = normalize(T_normalMatrix * vertexTangent.xyz);
	vec3 b = tangent.w * cross(outNormal, t);
	// t = normalize(t - (dot(t, outNormal) * outNormal)); //Ortho dinges
	tbnMatrix = mat3(t, b, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
vec2 CalcParallaxTexCoords(sampler2D dispMap, mat3 tbnMatrix, vec3 directionToEye, vec2 texCoords, float scale, float bias)
{
	return texCoords + (directionToEye * tbnMatrix).xy * (texture2D(dispMap, texCoords).r * scale + bias);
}
uniform sampler2D dispTexture;
uniform float dispScale;
uniform float dispBias;
uniform vec3 T_eyePos;
varying vec3 outNormal;
varying mat3 tbnMatrix;
uniform sampler2D normalTexture;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec3 directionToEye = normalize(T_eyePos - outPosition);
	vec2 uvCoords = outUV;
	uvCoords = CalcParallaxTexCoords(dispTexture, tbnMatrix, directionToEye, outUV, dispScale, dispBias);
	distortedTexCoords = uvCoords;
	vec3 normalVector = normalize(tbnMatrix * (255.0 / 128.0 * texture2D(normalTexture, distortedTexCoords).xyz - 1.0));
	vec3 fresnelNormalVector = normalize(outNormal);
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "bones cubeRefraction environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
#code "cubeRefraction environmentMapping"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
uniform vec3 T_eyePos;
uniform sampler2D R_refractionTexture;
uniform float refractiveIndex;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D environmentTexture;
uniform vec4 environmentColor;
uniform vec4 mixColor;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec2 distortedTexCoords = vec2(0.0);
	float alpha = 1.0;
	vec2 uvCoords = outUV;
	distortedTexCoords = uvCoords;
	vec3 normalVector = vec3(0.0,1.0,0.0);
	normalVector = normalize(outNormal);
	vec3 fresnelNormalVector = normalVector;
	vec3 eyeVector = normalize(T_eyePos - outPosition);
	vec3 reflectionColor = vec3(1.0);
	vec3 refractionColor = vec3(1.0);
	float refractionFactor = 1.0;
	vec3 refr = refract(-eyeVector, normalVector, refractiveIndex);
	if(length(refr) < 0.01) refractionFactor = 0.0;
	refractionColor = texture2D(R_refractionTexture, GetCubeTC(refr,R_cubeEdgeSize,R_cubeEdgeSize)).rgb;
	vec3 combinedColor = reflectionColor * refractionColor;
	vec4 textureColor = texture2D(environmentTexture, distortedTexCoords);
	vec3 finalColor = mix(combinedColor, mixColor.rgb, mixColor.a);
	gl_FragColor = vec4(finalColor, 1.0) * environmentColor * textureColor * alpha;
}

#endcode
